#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import Commands
import Dialogs
import QUB
import cStringIO
from math import *
import random
import urllib
import webbrowser

MAX_ITER = 100

profile = QUB.Profiles['K-Means'].active
if profile['weights'].data.count == 0:
    profile['weights'].data = 'weight'
if profile['useWeights'].data.count == 0:
    profile['useWeights'].data = 0
if profile['K'].data.count == 0:
    profile['K'].data = 3
if profile['trials'].data.count == 0:
    profile['trials'].data = 100
if profile['allK'].data.count == 0:
    profile['allK'].data = 0
if profile['normalize'].data.count == 0:
    profile['normalize'].data = 0

def getUsedNonzeroColors(result):
    colors = result.usedColors
    try:
        colors.remove(0)
    except:
        pass
    return colors

def min_index(lst):
    mni = 0
    for i in range(1,len(lst)):
        if lst[i] < lst[mni]:
            mni = i
    return mni

lastIter = 0
lastScore = 0.0

def DoKMeans(maxiter, inWeights=None, Report=QUB.Report, normalize=False):
    global lastIter, lastScore
    
    result = QUB.ResultsView.currResults
    colors = getUsedNonzeroColors(result)
    vars = [result[varname] for varname in result.visibleVars]
    segs = range(0, result.segmentCount)
    weights = inWeights or ([1.0] * result.segmentCount)

    if normalize:
        scales = [1.0 / (sum([var[i] for i in xrange(len(segs))]) / len(segs)) for var in vars]
    else:
        scales = [1.0 for var in vars]

    indexOfColor = {0:-1}
    for ix, color in enumerate(colors):
        indexOfColor[color] = ix

    def getCentroid(var, color):
        sumv = 0.0
        sumw = 0.0
        csegs = result.segmentsOfColor(color)
        for seg in csegs:
            sumv += weights[seg] * var[seg]
            sumw += weights[seg]
        if sumw == 0.0:
            return 0.0 # (0.0, 0.001, len(csegs)) # no points return whatever
        mean = sumv / sumw
        return mean
        #sumd = 0.0
        #for seg in csegs:
        #    sumd += weights[seg] * (var[seg] - mean)**2
        #std = sqrt(sumd / sumw) or (1.0e-3 * abs(mean)) or 1.0e-6  # change 0.0 std.dev to whatever the mean is
        #return (mean, std, len(csegs))        
    def getCentroids():
        ccc = [ [getCentroid(var, color) for var in vars]
                for color in colors ]
        #for cc in ccc:                     # change 0.0 std.dev to whatever the mean is,
        #    for i in range(len(cc)):       # unless the mean is also 0.0,
        #        if cc[i][1] == 0.0:        # then just use a nonsense constant
        #            cc[i] = (cc[i][0], cc[i][0], cc[i][2])
        #        if cc[i][1] == 0.0:
        #            cc[i] = (cc[i][0], 0.001, cc[i][2])
        #        if cc[i][2] == 0:
        #            cc[i] = (cc[i][0], cc[i][1], 1)
        return ccc
    def getMeasure(centroids):
        measPerSegPerColor = []
        for seg in segs:
            measPerColor = []
            for color, ccents in zip(colors, centroids):
                # K-Means requires a good concept of distance.  Euclidean distance on R^{nvar} is problematic
                # when one variable is orders of magnitude larger; it dominates.
                # We optionally scale (divide) each dimension by the mean coordinate.
                meas = 0.0
                for var, ccent, scale in zip(vars, ccents, scales):
                    meas += (scale * (var[seg] - ccent))**2
                measPerColor.append(weights[seg] * meas)
                #measPerColor.append(meas)
            measPerColor.append(1.0e20) # huge (bad) score for black coloration
            measPerSegPerColor.append( measPerColor )
        #QUB.Report(measPerSegPerColor)
        return measPerSegPerColor
    def getPicks(measPerSegPerColor):
        return [min_index(meas) for meas in measPerSegPerColor]
    def getScore(measPerSegPerColor, picks):
        #QUB.Report([meas[p] for meas,p in zip(measPerSegPerColor, picks)])
        return sum([meas[p] for meas,p in zip(measPerSegPerColor, picks)]) / len(picks)
    def getColors(picks):
        return [colors[maxi] for maxi in picks]
    def setColors(picks):
        result.doUpdates(0)
        for seg, pick in enumerate(picks):
            result.setSegmentColor(seg, colors[pick])
        result.doUpdates(1)

    centroids = getCentroids()
    meas = getMeasure(centroids)
    lastPicks = [indexOfColor[result.segmentColor(seg)] for seg in segs]
    lastScore = getScore(meas, lastPicks)
    lastIter = 0
    
    for iter in xrange(maxiter):
        picks = getPicks(meas)
        score = getScore(meas, picks)
        if score >= lastScore:
            break
        setColors(picks)
        centroids = getCentroids()
        meas = getMeasure(centroids)
        lastScore = score
        lastIter = iter+1
        
    Report("K-Means: %d iterations, %s mean deviation: %f" % (lastIter, normalize and 'normalized ' or '', lastScore))

def SelectKMeans():
    DoKMeans(MAX_ITER, normalize=(len(QUB.ResultsView.currResults.visibleVars)>1))

Commands.Register('Select', 'K-Means', SelectKMeans)

def SelectOneKMeans():
    DoKMeans(1, normalize=(len(QUB.ResultsView.currResults.visibleVars)>1))

Commands.Register('Select', 'K-Means one step', SelectOneKMeans)

def WeightedKMeans():
    weights = str(profile['weights'].data)
    result = QUB.ResultsView.currResults
    vars = result.keys()
    while True:
        userSays = QUB.Dialog('Which variable contains weights?', ['OK', 'Cancel'],
                              [Dialogs.DialogText('weights', '"select" variable name:', weights)])
        if userSays['button'] == 'Cancel':
            return
        weights = userSays['weights']
        if weights in vars:
            DoKMeans(MAX_ITER, result[weights], normalize=(len(QUB.ResultsView.currResults.visibleVars)>1))
            profile['weights'].data = weights
            return

Commands.Register('Select', 'K-Means weighted', WeightedKMeans)

def RandomK(symbols, count):
    seq = [0] * count
    used = {-1:1}
    for sym in symbols:
        pos = -1
        while used.has_key(pos):
            pos = random.randint(0,count-1)
        used[pos] = 1
        seq[pos] = sym
    return seq

def CountColors(colors):
    used = {}
    for c in colors:
        if c: # exckude black from the count
            used[c] = 1
    return len(used.keys())

def CanonicalColors(colors): # where i<j -> first i before first j
    renumber = {}
    nextNum = 1
    for c in colors:
        if not renumber.has_key(c):
            renumber[c] = nextNum
            nextNum += 1
    return [renumber[c] for c in colors]

def SelectXMeans():
    weights = str(profile['weights'].data)
    useWeights = profile['useWeights'].data[0]
    K = profile['K'].data[0]
    trials = profile['trials'].data[0]
    allK = profile['allK'].data[0]
    normalize = profile['normalize'].data[0]
    result = QUB.ResultsView.currResults
    vars = result.keys()
    goodInputs = False
    while not goodInputs:
        userSays = QUB.Dialog('X-Means', ['OK', 'Cancel'],
                              [Dialogs.DialogCheckBox('useWeights', 'Weighted', useWeights),
                               Dialogs.DialogText('weights', '"select" variable for weights:', weights),
                               Dialogs.DialogText('K', 'K:', K, Dialogs.acceptIntBetween(1,100)),
                               Dialogs.DialogCheckBox('allK', 'Test all k <= K', allK),
                               Dialogs.DialogText('trials', 'Trials:', trials, Dialogs.acceptIntGreaterThan(0)),
                               Dialogs.DialogCheckBox('normalize', 'Normalize each var to its mean', normalize)])
        if userSays['button'] == 'Cancel':
            return
        useWeights = userSays['useWeights']
        weights = userSays['weights']
        K = userSays['K']
        trials = userSays['trials']
        allK = userSays['allK']
        normalize = userSays['normalize']
        if useWeights and not (weights in vars):
            continue
        profile['weights'].data = weights
        profile['useWeights'].data = useWeights
        profile['K'].data = K
        profile['trials'].data = trials
        profile['allK'].data = allK
        profile['normalize'].data = normalize
        goodInputs = True

    nseg = result.segmentCount
    bestScore = 1.0e20
    bestColors = []
    Ks = [K]
    if allK:
        Ks = range(1,K+1)

    if K > nseg:
        QUB.Report('more colors than datapoints!')
        return    

    # tricky changing the meaning of K to loop var:
    for K in Ks:
        for trial in xrange((K>1) and trials or 1):
            colors = [0]*nseg
            while CountColors(colors) < K:
                colors = RandomK(range(1,K+1), nseg) # randomize one point each starting color 1..K
            result.doUpdates(0)
            for i, c in enumerate(colors):
                result.setSegmentColor(i, c)
            result.doUpdates(1)
            DoKMeans(MAX_ITER, result[weights], lambda x:x, normalize)
            DoKMeans(1, result[weights], lambda x:x, normalize)          # stabilize norm SSD
            colors = CanonicalColors([result.segmentColor(i) for i in xrange(nseg)])
            # QUB.Report("%f: %s" % (lastScore, colors))
            if (lastScore < bestScore) and (CountColors(colors) == K):
                bestScore = lastScore
                bestColors = colors
        QUB.Report("%d: %f: %s" % (K, bestScore, bestColors))
        result.doUpdates(0)
        for i, c in enumerate(bestColors):
            result.setSegmentColor(i, c)
        result.doUpdates(1)
    
Commands.Register('Select', 'X-Means', SelectXMeans)


def WebXMeans():
    result = QUB.ResultsView.currResults
    vars = [result[varname] for varname in result.visibleVars]
    
    data = cStringIO.StringIO()
    data.write('\t'.join(result.visibleVars))
    data.write('\n')
    for i in range(result.segmentCount):
        data.write('\t'.join([str(v[i]) for v in vars]))
        data.write('\n')
    webbrowser.open('http://www.qub.buffalo.edu/xmeans/index.cgi?sdata=%s' % urllib.quote(data.getvalue()))

Commands.Register('Select', 'X-Means (web)', WebXMeans)
